import json
import ntpath
from modeci_mdf.functions.standard import mdf_functions, create_python_expression
from typing import List, Tuple, Dict, Optional, Set, Any, Union
from modeci_mdf.utils import load_mdf, print_summary
from modeci_mdf.execution_engine import EvaluableGraph
import glom

expression_items = ["+", "*", "-", "/", "%", "(", ")"]


def convert_states_to_stateful_parameters(file_path: str = None, dt=5e-05):

    """Translates json file if with states to json file with stateful_parameters, otherwise unchanged
    Args:
            file_path: File in Json Format
    Returns:
            file in json format
    """
    f = open(file_path)
    data = json.load(f)

    filtered_list = [
        "parameters",
        "metadata",
        "functions",
        "states",
        "output_ports",
        "input_ports",
        "notes",
    ]
    all_nodes = []
    all_keys = []

    def keysExtractor(nested_dictionary):
        for k, v in nested_dictionary.items():

            if isinstance(v, dict) and k in "conditions":
                continue
            elif isinstance(v, dict):
                all_keys.append(k)
                if isinstance(v, dict) and k in "nodes":
                    break
                elif isinstance(v, dict):
                    keysExtractor(v)

    keysExtractor(data)
    path = ".".join(all_keys)

    def nodeExtractor(nested_dictionary: Dict[str, Any] = None):
        """Extracts all the node objects in the graph
        Args:
                nested_dictionary: input data
        Returns:
                Dictionary of node objects
        """
        for k, v in nested_dictionary.items():
            if isinstance(v, dict) and k in "nodes":
                all_nodes.append(v.keys())
            elif isinstance(v, dict):
                nodeExtractor(v)

    nodeExtractor(data)
    nodes_dict = dict.fromkeys(all_nodes[0])

    for key in list(nodes_dict.keys()):
        nodes_dict[key] = {}

    def parameterExtractor(nested_dictionary: Dict[str, Any] = None):
        """Extracts Parameters, states, functions, input and output ports at each node object
        Args:
                nested_dictionary: Input Data
        Returns:
                stores states, parameters, functions, input and output ports
        """
        for k, v in nested_dictionary.items():
            if isinstance(v, dict) and k in list(nodes_dict.keys()):
                for kk, vv in v.items():

                    if (isinstance(vv, dict) and kk in filtered_list) or (
                        isinstance(vv, str) and kk in filtered_list
                    ):

                        nodes_dict[k][kk] = vv
            if isinstance(v, dict):
                parameterExtractor(v)

    parameterExtractor(data)

    arg_dict = {}

    def get_arguments(d: Dict[str, Any] = None):
        """Extracts all parameters including stateful,dt for each node object
        Args:
                d: Node level dictionary with filtered keys
        Returns:
                all parameters for each node object
        """
        for key in d.keys():
            vi = []
            flag = 0
            if "parameters" in d[key].keys():
                vi += list(d[key]["parameters"].keys())
                for param in d[key]["parameters"].keys():
                    if "time_derivative" in d[key]["parameters"][param].keys():
                        flag = 1
                    if flag == 1 and "dt" not in vi:
                        vi.append("dt")

            if "states" in d[key].keys():
                vi += list(d[key]["states"].keys())
            # if 'states' in d[key].keys():
            # 	for state in d[key]['states'].keys():
            # 		if 'time_derivative' in d[key]['states'][state].keys():
            # 			flag = 1
            # 		if flag == 1 and 'dt' not in vi :
            # 			vi.append('dt')

            arg_dict[key] = vi

    get_arguments(nodes_dict)

    expression_dict = {}

    def get_expression(d: Dict[str, Any] = None):
        """get any expression (including time derivative) for each state or output port variable
        Args:
                d: Node level dictionary with filtered keys
        Returns:
                store expression for each state or output port variable
        """
        for key in d.keys():
            vi = []
            li = []
            temp_dic = {}
            if "parameters" in d[key].keys():
                for param in d[key]["parameters"].keys():

                    if "time_derivative" in d[key]["parameters"][param].keys():
                        li.append(param)
                        vi.append(d[key]["parameters"][param]["time_derivative"])
                    elif "value" in d[key]["parameters"][param].keys():

                        if isinstance(d[key]["parameters"][param]["value"], str):
                            if any(
                                x in d[key]["parameters"][param]["value"]
                                for x in expression_items
                            ):
                                li.append(param)
                                vi.append(d[key]["parameters"][param]["value"])
                            else:
                                li.append(param)
                                vi.append(None)
                        else:
                            li.append(param)
                            vi.append(None)
                    else:
                        li.append(param)
                        vi.append(None)

            if "output_ports" in d[key].keys():
                for output_port in d[key]["output_ports"].keys():
                    if isinstance(d[key]["output_ports"][output_port]["value"], str):
                        if any(
                            x in d[key]["output_ports"][output_port]["value"]
                            for x in expression_items
                        ):
                            li.append(output_port)
                            vi.append(d[key]["output_ports"][output_port]["value"])
                        else:
                            li.append(output_port)
                            vi.append(None)
            for i in range(len(vi)):
                temp_dic[li[i]] = vi[i]
            expression_dict[key] = temp_dic

    get_expression(nodes_dict)
    # print("expression_dict>>>", expression_dict)

    def createFunctions(d: Dict[str, Any] = None):
        """create functions for time_derivative expression for each state variable
        Args:
                d: Node level dictionary with filtered keys
        Returns:
                functions replacing time derivative for each state variable
        """
        for key in d.keys():

            if "functions" not in d[key].keys():
                d[key]["functions"] = {}

            if "parameters" in d[key].keys():
                parameterlist = []

                for idx, param in enumerate(list(d[key]["parameters"].keys())):
                    if "time_derivative" in d[key]["parameters"][param].keys():

                        d[key]["functions"][f"evaluated_{key}_{param}_next_value"] = {}
                        d[key]["functions"][f"evaluated_{key}_{param}_next_value"][
                            "value"
                        ] = {}

                        d[key]["functions"][f"evaluated_{key}_{param}_next_value"][
                            "value"
                        ] = (
                            str(param) + "+"
                            "(dt*" + str(expression_dict[key][param]) + ")"
                        )

                        # d[key]['functions']["evaluated_{}_{}_next_value".format(key, param)]['args']=  {}
                        # for pp in arg_dict[key]:
                        # 	d[key]['functions']["evaluated_{}_{}_next_value".format(key, param)]['args'][pp] = pp

                        parameterlist.append(param)
                    elif "value" in d[key]["parameters"][param].keys():
                        if isinstance(d[key]["parameters"][param]["value"], str):

                            if any(
                                x in d[key]["parameters"][param]["value"]
                                for x in expression_items
                            ):
                                d[key]["functions"][
                                    f"evaluated_{key}_{param}_next_value"
                                ] = {}
                                d[key]["functions"][
                                    f"evaluated_{key}_{param}_next_value"
                                ]["value"] = {}

                                d[key]["functions"][
                                    f"evaluated_{key}_{param}_next_value"
                                ]["value"] = expression_dict[key][param]
                                # d[key]['functions']["evaluated_{}_{}_next_value".format(key, param)]['args']=  {}
                                # for pp in arg_dict[key]:
                                # d[key]['functions']["evaluated_{}_{}_next_value".format(key, param)]['args'][pp] = pp
                                parameterlist.append(param)

                    elif "function" in d[key]["parameters"][param].keys():

                        d[key]["functions"][param] = {}
                        d[key]["functions"][param]["function"] = {}

                        d[key]["functions"][param]["function"][
                            d[key]["parameters"][param]["function"]
                        ] = {}
                        d[key]["functions"][param]["function"][
                            d[key]["parameters"][param]["function"]
                        ] = d[key]["parameters"][param]["args"]

                    if idx > 0:
                        for prev_param in parameterlist[:-1]:

                            # d[key]['functions']["evaluated_{}_{}_next_value".format(key, param)]['args'][prev_param] = "evaluated_{}_{}_next_value".format(key, prev_param)
                            d[key]["functions"][f"evaluated_{key}_{param}_next_value"][
                                "value"
                            ] = d[key]["functions"][
                                f"evaluated_{key}_{param}_next_value"
                            ][
                                "value"
                            ].replace(
                                prev_param, f"evaluated_{key}_{prev_param}_next_value"
                            )

            if "output_ports" in d[key].keys():

                for idx, output_port in enumerate(list(d[key]["output_ports"].keys())):
                    if isinstance(d[key]["output_ports"][output_port]["value"], str):
                        if any(
                            x in d[key]["output_ports"][output_port]["value"]
                            for x in expression_items
                        ):
                            d[key]["functions"][
                                f"evaluated_{key}_{output_port}_value"
                            ] = {}
                            d[key]["functions"][f"evaluated_{key}_{output_port}_value"][
                                "value"
                            ] = {}
                            d[key]["functions"][f"evaluated_{key}_{output_port}_value"][
                                "value"
                            ] = expression_dict[key][output_port]
                            # d[key]['functions']["evaluated_{}_{}_value".format(key, output_port)]['args']=  {}
                            # for param in arg_dict[key]:
                            # 	d[key]['functions']["evaluated_{}_{}_value".format(key, output_port)]['args'][param] = param

    createFunctions(nodes_dict)

    def changetoValue(d: Dict[str, Any] = None):
        """Converts states into stateful_parameters, adds dt to parameters
        Args:
                d: dictionary with states information at the Node
        Returns:
                dictionary with stateful_parameters information
        """
        for key in d.keys():
            if "parameters" in d[key].keys():
                for param in list(d[key]["parameters"].keys()):
                    if "time_derivative" in d[key]["parameters"][param].keys():
                        d[key]["parameters"]["dt"] = {}
                        d[key]["parameters"]["dt"]["value"] = dt
                        if (
                            "default_initial_value"
                            in d[key]["parameters"][param].keys()
                        ):
                            if (
                                d[key]["parameters"][param]["default_initial_value"]
                                in d[key]["parameters"].keys()
                            ):
                                d[key]["parameters"][param][
                                    "default_initial_value"
                                ] = d[key]["parameters"][
                                    d[key]["parameters"][param]["default_initial_value"]
                                ][
                                    "value"
                                ]

                        else:
                            d[key]["parameters"][param]["default_initial_value"] = 0

                        d[key]["parameters"][param].pop("time_derivative")
                        d[key]["parameters"][param][
                            "value"
                        ] = f"evaluated_{key}_{param}_next_value"
                        d[key]["parameters"]["time"] = {
                            "default_initial_value": 0,
                            "value": "evaluated_time_next_value",
                        }

                        d[key]["functions"]["evaluated_time_next_value"] = {}
                        d[key]["functions"]["evaluated_time_next_value"][
                            "function"
                        ] = {}

                        d[key]["functions"]["evaluated_time_next_value"]["function"][
                            "linear"
                        ] = {}

                        d[key]["functions"]["evaluated_time_next_value"]["function"][
                            "linear"
                        ] = {"variable0": "time", "slope": 1, "intercept": "dt"}

                    elif "value" in d[key]["parameters"][param].keys():
                        if isinstance(d[key]["parameters"][param]["value"], str):
                            if any(
                                x in d[key]["parameters"][param]["value"]
                                for x in expression_items
                            ):
                                if (
                                    "default_initial_value"
                                    in d[key]["parameters"][param].keys()
                                ):
                                    if (
                                        d[key]["parameters"][param][
                                            "default_initial_value"
                                        ]
                                        in d[key]["parameters"].keys()
                                    ):
                                        d[key]["parameters"][param][
                                            "default_initial_value"
                                        ] = d[key]["parameters"][
                                            d[key]["parameters"][param][
                                                "default_initial_value"
                                            ]
                                        ][
                                            "value"
                                        ]

                                else:
                                    d[key]["parameters"][param][
                                        "default_initial_value"
                                    ] = 0

                                d[key]["parameters"][param][
                                    "value"
                                ] = f"evaluated_{key}_{param}_next_value"

                    elif "function" in d[key]["parameters"][param].keys():
                        d[key]["parameters"].pop(param)

            if "output_ports" in d[key].keys():
                for output_port in list(d[key]["output_ports"].keys()):
                    if isinstance(d[key]["output_ports"][output_port]["value"], str):
                        if any(
                            x in d[key]["output_ports"][output_port]["value"]
                            for x in expression_items
                        ):

                            d[key]["output_ports"][output_port][
                                "value"
                            ] = f"evaluated_{key}_{output_port}_value"

        for key in d.keys():
            if "states" in d[key].keys():
                d[key]["stateful_parameters"] = d[key].pop("states")

    changetoValue(nodes_dict)
    glom.assign(data, path, nodes_dict)

    def repl(dr):
        """Replaces all names containing states into stateful_parameters
        Args:
                dr: full dictionary with states name
        Returns:
                dictionary with stateful_parameters name
        """
        dr = str(dr)
        dr = dr.replace("states", "stateful_parameters")
        dr = dr.replace("States", "Stateful_Parameters")
        dr = dr.replace("state_example", "stateful_parameters_example")
        return eval(dr)

    data = repl(data)
    return data



"""
    Useful utility functions for dealing with MDF objects.
"""

from modeci_mdf.mdf import Model, Graph, Node, Edge, OutputPort, Function, InputPort


def create_example_node(node_id: str, graph: Graph) -> Node:
    """
    Create a simple example node with Input inside a graph

    Args:
        node_id: The unique id for the first node in the graph.
        graph: The graph to add the example node.

    Returns:
        The node (with id=node_id) created in the graph.
    """

    a = Node(id=node_id)
    graph.nodes.append(a)

    a.parameters = {"logistic_gain": 3, "slope": 0.5, "intercept": 0}
    ip1 = InputPort(id="input_port1", shape="(1,)")
    a.input_ports.append(ip1)

    f1 = Function(
        id="logistic_1",
        function="logistic",
        args={"variable0": ip1.id, "gain": "logistic_gain", "bias": 0, "offset": 0},
    )
    a.functions.append(f1)
    f2 = Function(
        id="linear_1",
        function="linear",
        args={"variable0": f1.id, "slope": "slope", "intercept": "intercept"},
    )
    a.functions.append(f2)
    a.output_ports.append(OutputPort(id="output_1", value="linear_1"))

    return a


def simple_connect(pre_node, post_node, graph) -> Edge:
    """
    Create an edge connecting two nodes in a graph.

    Args:
        pre_node: The source node.
        post_node: The destination node.
        graph: The graph to and the edge.

    Returns:
        The edge that has been added to the graph.
    """

    e1 = Edge(
        id=f"edge_{pre_node.id}_{post_node.id}",
        sender=pre_node.id,
        sender_port=pre_node.output_ports[0].id,
        receiver=post_node.id,
        receiver_port=post_node.input_ports[0].id,
    )

    graph.edges.append(e1)
    return e1


def print_summary(graph: Graph):
    """Print a summary of a graph to standard out."""
    print(
        "Graph %s with %i nodes and %s edges\n"
        % (graph.id, len(graph.nodes), len(graph.edges))
    )
    for node in graph.nodes:
        print("%s" % node)
    for edge in graph.edges:
        print("%s" % edge)


def load_mdf(filename: str) -> Model:
    """
    Load an MDF file from JSON or YAML. File type is detected automatically based on extension.
    """
    return Model.from_file(filename)


def load_mdf_json(filename: str) -> Model:
    """
    Load an MDF JSON file
    """
    return Model.from_json_file(filename)


def load_mdf_yaml(filename: str) -> Model:
    """
    Load an MDF YAML file
    """
    return Model.from_yaml_file(filename)


def color_rgb_to_hex(rgb):
    """Convert a rgb color to hexadecimal format."""
    color = "#"
    print("Converting %s to hex color" % rgb)
    for a in rgb.split():
        color = color + "%02x" % int(float(a) * 255)
    return color


def is_number(s):
    """Return :code:`True` if cast to :code:`float` does not throw ValueError, :code:`False` otherwise."""
    try:
        float(s)
        return True
    except (TypeError, ValueError):
        return False



"""
Code for importing ONNX models into MDF.
"""
import typing

import onnx

from onnx import (
    ModelProto,
    TensorProto,
    GraphProto,
    AttributeProto,
    numpy_helper,
    shape_inference,
)
from onnx.defs import get_schema

from modeci_mdf.mdf import *


def id_to_port(id: str):
    """Turn unique ONNX output and input value names into valid MDF input and outport names"""

    new_name = str(id).replace(".", "_")

    # Get rid of double colon in id names, this causes issues with execution engine.
    new_name = new_name.replace("::", "_")

    # If the first character is a digit, precede with an underscore so this can never be interpreted
    # as number down the line.
    if new_name[0].isdigit():
        new_name = "_" + new_name

    return new_name


def get_shape_params(shape: onnx.TensorShapeProto) -> typing.Tuple:
    """
    Small helper function to extract a tuple from the TensorShapeProto. These objects
    can contain both integer dimensions and parameter dimensions that are variable, like
    'batch_size'.

    Args:
        shape: The ONNX shape proto to process.

    Returns:
        A tuple that can contain both integers and strings for parameter dimensions.
    """
    shape = tuple(d.dim_param if d.dim_param != "" else d.dim_value for d in shape.dim)

    # If shape is empty tuple, its a scalar, make it size 1
    if len(shape) == 0:
        shape = (1,)

    return shape


def get_onnx_attribute(a):

    # Use the helpers to get the appropriate value
    val = onnx.helper.get_attribute_value(a)

    # get_attribute_value() can return TensorProto's, lets convert them to a list for JSON
    # FIXME: This begs the question, is JSON a good format for storing large tensors (nope)
    if type(val) == TensorProto:
        return numpy_helper.to_array(val).tolist()
    else:
        return val


def onnx_node_to_mdf(
    node: typing.Union[onnx.NodeProto, onnx.ValueInfoProto],
    onnx_initializer: typing.Dict[str, typing.Dict[str, typing.Any]],
) -> Node:
    """
    Construct an MDF node (and function) from an ONNX NodeProto or ValueInfoProto

    Args:
        node: The ONNX node to use to form the MDF node. Can be a node from the model or
            a ValueInfoProto specifying an input or output.
        onnx_initializer: A specification of values in the graph that ONNX has
            marked as initializer's. This dict is keyed on the name of the parameter,
            the value is another dict with three entries; shape, type, and value.

    Returns:
        The equivalent MDF node for the ONNX node passed in as argument.
    """

    # If this is a ONNX Node,
    if type(node) == onnx.NodeProto:

        # Create and MDF node with parameters
        # FIXME: We need to preserve type info somewhere
        params_dict = {a.name: get_onnx_attribute(a) for a in node.attribute}

        # For any attributes that are sub-graphs, we need to recurse
        for aname, val in params_dict.items():
            if type(val) == GraphProto:
                params_dict[aname] = onnx_to_mdf(val, onnx_initializer=onnx_initializer)

        # If we have we have value constants that feed into this node. Make them parameters
        # instead of input ports
        non_constant_inputs = []
        func_args = {}
        for inp_i, inp in enumerate(node.input):

            # Get the name of the formal argument that corresponds to this input.
            # We need to go to the schema for this.
            # FIXME: We need to make sure we are going the correct schema here ... yuck!
            try:
                arg_name = get_schema(node.op_type).inputs[inp_i].name
            except IndexError:
                arg_name = f"arg_{inp}"

            if inp in onnx_initializer and "value" in onnx_initializer[inp]:
                params_dict[arg_name] = onnx_initializer[inp]["value"]
                func_args[arg_name] = arg_name
            else:
                non_constant_inputs.append(inp)
                func_args[arg_name] = id_to_port(inp)

        # FIXME: parameters must be set or we get JSON serialization error later
        mdf_node = Node(id=node.name)

        for p in params_dict:
            if type(params_dict[p]) == Graph:
                mdf_node.parameters.append(
                    Parameter(
                        id=p, value={"graph_%s" % params_dict[p].id: params_dict[p]}
                    )
                )
            else:
                mdf_node.parameters.append(Parameter(id=p, value=params_dict[p]))

        # Add the function
        # FIXME: There is probably more stuff we need to preserve for ONNX Ops
        func = Parameter(id=node.name, function=f"onnx::{node.op_type}", args=func_args)
        mdf_node.parameters.append(func)

        # Recreate inputs and outputs of ONNX node as InputPorts and OutputPorts
        for inp in non_constant_inputs:
            param_info = onnx_initializer.get(inp, None)
            shape = param_info["shape"] if param_info else None
            ip = InputPort(id=id_to_port(inp), shape=shape)
            mdf_node.input_ports.append(ip)

        for out in node.output:
            op = OutputPort(id=id_to_port(out), value=func.id)
            mdf_node.output_ports.append(op)

    elif type(node) == onnx.ValueInfoProto:
        raise NotImplementedError()
        # # Lets start with an MDF node that uses the ONNX node name as its id. No parameters
        # mdf_node = Node(id=node.name)
        #
        # # This is an input or output node. No Op\Function or parameters. This is just
        # # a simple pass through node with an input and output port with the correct
        # # shape.
        # # FIXME: Should this be necessary? ONNX treats input and output nodes as simple named values.
        # ip1 = InputPort(id=f"in_port",
        #                 shape=str(get_shape_params(node.type.tensor_type.shape))) # FIXME: Why string?
        # mdf_node.input_ports.append(ip1)
        # op1 = OutputPort(id=node.name)
        # op1.value = f"in_port"
        # mdf_node.output_ports.append(op1)

    return mdf_node


def onnx_to_mdf(
    onnx_model: typing.Union[ModelProto, GraphProto],
    onnx_initializer: typing.Dict[str, typing.Dict[str, typing.Any]] = None,
):
    """
    Convert a loaded ONNX model into a MDF model.

    Args:
        onnx_model: The ONNX model to convert. Typically, this is the result of a call to onnx.load()
        onnx_initializer: A specification of values in the graph that ONNX has
            marked as initializer's. This dict is keyed on the name of the parameter,
            the value is another dict with three entries; shape, type, and value.

    Returns:
        An MDF description of the ONNX model.
    """

    if onnx_initializer is None:
        onnx_initializer = {}

    if type(onnx_model) == ModelProto:

        # Do shape inference on the model so we can get shapes of intermediate outputs
        # FIXME: This function has side-effects, it probably shouldn't
        try:
            onnx_model = shape_inference.infer_shapes(onnx_model)
        except RuntimeError:
            pass

        graph = onnx_model.graph

    else:
        graph = onnx_model

    # Get all the nodes in the onnx model, even the inputs and outputs
    onnx_nodes = list(graph.node)

    if hasattr(graph, "initializer"):
        # Parameters that have been initialized with values.
        # FIXME: We need a cleaner way to extract this info.
        onnx_initializer_t = {}
        for t in graph.initializer:
            t_np = numpy_helper.to_array(t)
            onnx_initializer_t[t.name] = {"shape": t_np.shape, "type": str(t_np.dtype)}

        # And the input and intermediate node shapes as well
        for vinfo in list(graph.input) + list(graph.value_info):
            vshape = get_shape_params(vinfo.type.tensor_type.shape)

            try:
                vtype = onnx.helper.printable_type(vinfo.type)
            except AssertionError:
                # Couldn't extract type
                vtype = None

            onnx_initializer_t[vinfo.name] = {"shape": vshape, "type": vtype}

        onnx_initializer = {**onnx_initializer, **onnx_initializer_t}

    # Finally, some nodes are constants, extract the values and drop the nodes.
    # They will be removed in the MDF and passed as named parameters to the Node
    constants = {}
    onnx_nodes_nc = []
    for onnx_node in onnx_nodes:
        if onnx_node.op_type == "Constant":
            v = get_onnx_attribute(onnx_node.attribute[0])
            constants[onnx_node.output[0]] = {
                "shape": v.shape if hasattr(v, "shape") else (1,),
                "type": str(v.dtype) if hasattr(v, "dtype") else str(type(v)),
                "value": v,
            }
        else:
            onnx_nodes_nc.append(onnx_node)
    onnx_nodes = onnx_nodes_nc

    # Add constants to the initializer dict
    onnx_initializer = {**onnx_initializer, **constants}

    mod_graph = Graph(id=graph.name)

    # Construct the equivalent nodes in MDF
    mdf_nodes = [
        onnx_node_to_mdf(node=node, onnx_initializer=onnx_initializer)
        for node in onnx_nodes
    ]

    mod_graph.nodes.extend(mdf_nodes)

    # Construct the edges, we will do this by going through all the nodes.
    node_pairs = list(zip(onnx_nodes, mod_graph.nodes))
    for onnx_node, mdf_node in node_pairs:
        if len(onnx_node.output) > 0:
            for i, out in enumerate(onnx_node.output):
                out_port_id = mdf_node.output_ports[i].id

                # Find all node input ports with this outport id
                # FIXME: This is slow for big graphs with lots of edges. Best to build a data structure for this.
                receiver = [
                    (m, ip)
                    for n, m in node_pairs
                    for ip in m.input_ports
                    if out_port_id == ip.id
                ]

                # Make an edge for each receiver of this output port
                for receiver_node, receiver_port in receiver:
                    edge = Edge(
                        id=f"{mdf_node.id}.{out_port_id}_{receiver_node.id}.{receiver_port.id}",
                        sender=mdf_node.id,
                        sender_port=out_port_id,
                        receiver=receiver_node.id,
                        receiver_port=receiver_port.id,
                    )

                    mod_graph.edges.append(edge)

    # If they passed an ONNX model, wrap the graph in a MDF model
    if type(onnx_model) == ModelProto:
        mod = Model(id="ONNX Model")
        mod.graphs.append(mod_graph)
        return mod

    else:
        return mod_graph


def find_subgraphs(
    graph: onnx.GraphProto, graph_dict: typing.Dict[str, GraphProto] = None
) -> typing.Dict[str, GraphProto]:
    """
    Recurse through an ONNX graph and find all subgraphs.

    Args:
        graph: The graph to search.
        graph_list: Insert graphs we find into this dict. Use the parent node name as a key.
            If None, intitialize to empty dict.

    Returns:
        All the subgraphs in the for the graph.
    """

    if graph_dict is None:
        graph_dict = {}

    for node in graph.node:
        for ai, attr in enumerate(node.attribute):
            if attr.type == AttributeProto.GRAPH:
                subgraph = onnx.helper.get_attribute_value(attr)
                graph_dict[f"{node.name}_attr{ai}"] = subgraph
                graph_dict = find_subgraphs(subgraph, graph_dict)
            elif attr.type == AttributeProto.GRAPHS:
                subgraphs = onnx.helper.get_attribute_value(attr)
                for gi, subgraph in enumerate(subgraphs):
                    graph_dict[f"{node.name}_attr{ai}_g{gi}"] = subgraph
                    graph_dict = find_subgraphs(subgraph, graph_dict)

    return graph_dict


def convert_file(input_file: str):
    """
    Simple converter from ONNX to MDF. Takes in ONNX files and generates MDF JSON/YAML files.

    Args:
        input_file: The input file path to the ONNX file. Output files are generated in same
            directory with -mdf.json and -mdf.yml extensions.

    Returns:
        MoneType
    """

    import os

    out_filename = f"{os.path.splitext(input_file)[0]}-mdf"
    onnx_model = onnx.load(input_file)
    onnx.checker.check_model(onnx_model)
    mdf_model = onnx_to_mdf(onnx_model)
    mdf_model.to_json_file(f"{out_filename}.json")
    mdf_model.to_yaml_file(f"{out_filename}.yaml")


def main():

    import argparse

    parser = argparse.ArgumentParser(
        description="Simple converter from ONNX to MDF. "
        "Takes in ONNX files and generates MDF JSON/YAML"
    )
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_file",
        type=str,
        help="An input ONNX file. "
        "Output files are generated in same directory "
        "with -mdf.json and -mdf.yml extensions.",
    )

    args = parser.parse_args()

    convert_file(args.input_file)


if __name__ == "__main__":
    main()



"""
Simple export of MDF to NeuroML(2/lite) and LEMS.

Work in progress...

"""

import sys
import os
import neuromllite
import lems.api as lems

from modeci_mdf.functions.standard import mdf_functions, substitute_args
from modeci_mdf.execution_engine import evaluate_expr


def mdf_to_neuroml(graph, save_to=None, format=None, run_duration_sec=2):

    print("Converting graph: %s to NeuroML" % (graph.id))

    net = neuromllite.Network(id=graph.id)
    net.notes = "NeuroMLlite export of {} graph: {}".format(
        format if format else "MDF",
        graph.id,
    )

    model = lems.Model()
    lems_definitions = "%s_lems_definitions.xml" % graph.id

    for node in graph.nodes:
        print("    Node: %s" % node.id)

        node_comp_type = "%s__definition" % node.id
        node_comp = "%s__instance" % node.id

        # Create the ComponentType which defines behaviour of the general class
        ct = lems.ComponentType(node_comp_type, extends="baseCellMembPotDL")
        ct.add(lems.Attachments("only_input_port", "basePointCurrentDL"))
        ct.dynamics.add(
            lems.DerivedVariable(name="V", dimension="none", value="0", exposure="V")
        )
        model.add(ct)

        # Define the Component - an instance of the ComponentType
        comp = lems.Component(node_comp, node_comp_type)
        model.add(comp)

        cell = neuromllite.Cell(id=node_comp, lems_source_file=lems_definitions)
        net.cells.append(cell)

        pop = neuromllite.Population(
            id=node.id,
            size=1,
            component=cell.id,
            properties={"color": "0.2 0.2 0.2", "radius": 3},
        )
        net.populations.append(pop)

        if len(node.input_ports) > 1:
            raise Exception("Currently only max 1 input port supported in NeuroML...")

        for ip in node.input_ports:
            ct.add(lems.Exposure(ip.id, "none"))
            ct.dynamics.add(
                lems.DerivedVariable(
                    name=ip.id,
                    dimension="none",
                    select="only_input_port[*]/I",
                    reduce="add",
                    exposure=ip.id,
                )
            )

        on_start = None

        for p in node.parameters:
            print("Converting %s" % p)
            if p.value is not None:
                try:
                    v_num = float(p.value)
                    ct.add(lems.Parameter(p.id, "none"))
                    comp.parameters[p.id] = v_num
                    print(comp.parameters[p.id])
                except Exception as e:

                    ct.add(lems.Exposure(p.id, "none"))
                    dv = lems.DerivedVariable(
                        name=p.id,
                        dimension="none",
                        value="%s" % (p.value),
                        exposure=p.id,
                    )
                    ct.dynamics.add(dv)

            elif p.function is not None:
                ct.add(lems.Exposure(p.id, "none"))
                func_info = mdf_functions[p.function]
                expr = func_info["expression_string"]
                expr2 = substitute_args(expr, p.args)
                for arg in p.args:
                    expr += ";{}={}".format(arg, p.args[arg])
                dv = lems.DerivedVariable(
                    name=p.id, dimension="none", value="%s" % (expr2), exposure=p.id
                )
                ct.dynamics.add(dv)
            else:
                ct.add(lems.Exposure(p.id, "none"))
                ct.dynamics.add(
                    lems.StateVariable(name=p.id, dimension="none", exposure=p.id)
                )
                if p.default_initial_value:
                    if on_start is None:
                        on_start = lems.OnStart()
                        ct.dynamics.add(on_start)
                    sa = lems.StateAssignment(
                        variable=p.id, value=str(evaluate_expr(p.default_initial_value))
                    )
                    on_start.actions.append(sa)

                if p.time_derivative:
                    td = lems.TimeDerivative(variable=p.id, value=p.time_derivative)
                    ct.dynamics.add(td)

        if len(node.output_ports) > 1:
            raise Exception("Currently only max 1 output port supported in NeuroML...")

        for op in node.output_ports:
            ct.add(lems.Exposure(op.id, "none"))
            ct.dynamics.add(
                lems.DerivedVariable(
                    name=op.id, dimension="none", value=op.value, exposure=op.id
                )
            )
            only_output_port = "only_output_port"
            ct.add(lems.Exposure(only_output_port, "none"))
            ct.dynamics.add(
                lems.DerivedVariable(
                    name=only_output_port,
                    dimension="none",
                    value=op.id,
                    exposure=only_output_port,
                )
            )

    if len(graph.edges) > 0:

        model.add(
            lems.Include(os.path.join(os.path.dirname(__file__), "syn_definitions.xml"))
        )
        rsDL = neuromllite.Synapse(id="rsDL", lems_source_file=lems_definitions)
        net.synapses.append(rsDL)
        # syn_id = 'silentSyn'
        # silentSynDL = neuromllite.Synapse(id=syn_id, lems_source_file=lems_definitions)

    for edge in graph.edges:
        print(f"    Edge: {edge.id} connects {edge.sender} to {edge.receiver}")

        ssyn_id = "silentSyn_proj_%s" % edge.id
        ssyn_id = "silentSyn_proj_%s" % edge.id
        # ssyn_id = 'silentSynX'
        silentDLin = neuromllite.Synapse(id=ssyn_id, lems_source_file=lems_definitions)

        model.add(lems.Component(ssyn_id, "silentRateSynapseDL"))

        net.synapses.append(silentDLin)

        net.projections.append(
            neuromllite.Projection(
                id="proj_%s" % edge.id,
                presynaptic=edge.sender,
                postsynaptic=edge.receiver,
                synapse=rsDL.id,
                pre_synapse=silentDLin.id,
                type="continuousProjection",
                weight=1,
                random_connectivity=neuromllite.RandomConnectivity(probability=1),
            )
        )

    # Much more todo...
    model.export_to_file(lems_definitions)

    print("Nml net: %s" % net)
    if save_to:
        new_file = net.to_json_file(save_to)
        print("Saved NML to: %s" % save_to)

    ################################################################################
    ###   Build Simulation object & save as JSON

    simtime = 1000 * run_duration_sec
    dt = 0.1
    sim = neuromllite.Simulation(
        id="Sim%s" % net.id,
        network=new_file,
        duration=simtime,
        dt=dt,
        seed=123,
        record_variables={"OUTPUT": {"all": "*"}},
    )

    record_variables = {}
    for node in graph.nodes:
        for ip in node.input_ports:
            if not ip.id in record_variables:
                record_variables[ip.id] = {}
            record_variables[ip.id][node.id] = 0

        for p in node.parameters:
            if p.is_stateful():
                if not p.id in record_variables:
                    record_variables[p.id] = {}
                record_variables[p.id][node.id] = 0

        for op in node.output_ports:
            if not op.id in record_variables:
                record_variables[op.id] = {}
            record_variables[op.id][node.id] = 0

    sim.record_variables = record_variables
    if save_to:
        sf = sim.to_json_file()

        print("Saved Simulation to: %s" % sf)

    return net, sim


if __name__ == "__main__":

    from modeci_mdf.utils import load_mdf, print_summary

    example = "../../../../examples/MDF/Simple.json"
    verbose = True
    run = False
    if "-run" in sys.argv:
        run = True
        sys.argv.remove("-run")

    if len(sys.argv) >= 2:
        example = sys.argv[1]
        verbose = False

    model = load_mdf(example)
    mod_graph = model.graphs[0]

    print("Loaded Graph:")
    print_summary(mod_graph)

    print("------------------")
    nmllite_file = example.replace(".json", ".nmllite.json")
    net, sim = mdf_to_neuroml(
        mod_graph, save_to=nmllite_file, format=model.format, run_duration_sec=2
    )

    if run:
        sf = "%s.json" % sim.id
        print("Running the model: %s" % sf)
        from neuromllite.NetworkGenerator import generate_and_run

        simulator = "jNeuroML"
        traces, events = generate_and_run(
            sim,
            simulator,
            network=net,
            return_results=True,
            base_dir=None,
            target_dir=None,
            num_processors=1,
        )
        for t in traces:
            if t != "t":  # the time array
                print("    {} = {}".format(t, traces[t][-1]))



from .model import Model
from collections import UserDict


class Chunk(UserDict):
    def __init__(self, contents, bound=None):
        UserDict.__init__(self)
        if isinstance(contents, Chunk):
            self.update(contents)
        elif isinstance(contents, str):
            for i, x in enumerate(contents.split()):
                if ":" in x:
                    i, x = x.split(":", 1)
                if x.startswith("?"):
                    key = x[1:]
                    x = bound[key]
                self[i] = x
        elif hasattr(contents, "__dict__"):
            for k, v in contents.__dict__.items():
                if type(v) in [str, float, int, bool]:
                    self[k] = v
        else:
            try:
                self.update(contents)
            except:
                raise Exception("Unknown contents for chunk:", contents)

    def __repr__(self):
        r = []
        keys = self.keys()
        i = 0
        while i in keys:
            r.append("%s" % self[i])
            keys.remove(i)
            i += 1
        keys.sort()
        for k in keys:
            if k[0] != "_":
                r.append("{}:{}".format(k, self[k]))
        return " ".join(r)


class Buffer(Model):
    def __init__(self):
        self.chunk = None

    def set(self, chunk):
        try:
            self.chunk = Chunk(chunk, self.sch.bound)
        except AttributeError:
            self.chunk = Chunk(chunk, {})

    def modify(self, **args):
        for k, v in args.items():
            if k.startswith("_"):
                k = int(k[1:])
            if k not in self.chunk:
                raise Exception(f'No slot "{k}" to modify to "{v}"')
            self.chunk[k] = v
            self.chunk = self.chunk

    def __getitem__(self, key):
        return self.chunk[key]

    def clear(self):
        self.chunk = None

    def __eq__(self, other):
        return self.chunk == other

    def __hash__(self):
        return id(self)

    def __len__(self):
        if self.chunk is None:
            return 0
        return len(self.chunk)

    def isEmpty(self):
        return len(self) == 0



import heapq

from . import logger


class Trigger:
    def __init__(self, name=""):
        self.name = name

    def __str__(self):
        return '<Trigger "%s">' % self.name


class Event:
    generator = False

    def __init__(self, func, time, args=[], keys={}, priority=0):
        self.name = getattr(func, "func_name", None)

        try:
            code = func.func_code
        except AttributeError:
            try:
                code = func.__call__.im_func.func_code
            except:
                code = None
        if code and code.co_flags & 0x20 == 0x20:  # check for generator
            func = func(*args, **keys).next
            args = []
            keys = {}
            self.generator = True

        self.func = func
        self.args = args
        self.keys = keys
        self.time = time
        self.priority = priority
        self.group = ()
        self.cancelled = False
        self.parent = None

    def __cmp__(self, other):
        return cmp((self.time, -self.priority), (other.time, -other.priority))

    def __repr__(self):
        return "<{} {:x} {:5.3f}>".format(self.name, id(self.func), self.time)


class SchedulerError(Exception):
    pass


class Scheduler:
    def __init__(self):
        self.queue = []
        self.to_be_added = []
        self.triggers = {}
        self.time = 0.0
        self.stop_flag = False
        self.log = logger.log_proxy

    def extend(self, other):
        for k, v in other.triggers.items():
            if k not in self.triggers:
                self.triggers[k] = v
            else:
                self.triggers[k].extend(v)
        if len(other.queue) > 0:
            self.queue.extend(other.queue)
            heapq.heapify(self.queue)

    def trigger(self, key, priority=None):
        if key in self.triggers:
            for event in self.triggers[key]:
                event.time = self.time
                if priority is not None:
                    event.priority = priority
                self.add_event(event)
            del self.triggers[key][:]

    def add_event(self, event):
        heapq.heappush(self.queue, event)

    def add(self, func, delay=0, args=[], keys={}, priority=0, thread_safe=False):
        if thread_safe:
            self.to_be_added.append((func, delay, args, keys, priority))
        else:
            ev = Event(func, self.time + delay, args=args, keys=keys, priority=priority)
            self.add_event(ev)
            return ev

    def run(self):
        self.stop_flag = False
        while not self.stop_flag and len(self.queue) > 0:
            next = self.queue[0].time
            if next > self.time:
                self.time = next
                self.log.time = next
            self.do_event(heapq.heappop(self.queue))
            while self.to_be_added:
                self.add(*self.to_be_added.pop())

    def handle_result(self, result, event):
        if isinstance(result, (int, float)):
            event.time = self.time + result
            self.add_event(event)
        elif isinstance(result, dict):
            event.time = self.time + result.get("delay", 0)
            event.priority = result.get("priority", event.priority)
            self.add_event(event)
        elif isinstance(result, (str, Trigger)):
            event.time = None
            if result not in self.triggers:
                self.triggers[result] = [event]
            else:
                self.triggers[result].append(event)
        elif isinstance(result, (list, tuple)):
            events = [copy.copy(event) for r in result]
            for e in events:
                e.group = events
            for i, r in enumerate(result):
                self.handle_result(r, events[i])
        elif result is None:
            if event.parent is not None:
                event.parent.time = self.time
                self.add_event(event.parent)
        elif isinstance(result, Event):
            if result.generator and event.generator:
                result.parent = event
        elif hasattr(result, "default_trigger"):
            self.handle_result(result.default_trigger, event)
        else:
            raise SchedulerError("Incorrect 'yield': %s" % (result))

    def do_event(self, event):
        assert self.time == event.time

        if event.cancelled:
            return
        for e in event.group:
            e.cancelled = True
        event.cancelled = False

        try:
            result = event.func(*event.args, **event.keys)
        except StopIteration:
            result = None

        self.handle_result(result, event)

    def stop(self):
        self.stop_flag = True



"""
    Code for importing ACT-R models into MDF.
"""
from modeci_mdf.mdf import *


def build_model() -> Model:
    """Builds the base model of the ACT-R production system without any chunks,
    goals, or productions specified.

    Returns:
        An MDF model object representing the core ACT-R production system.
    """
    mod = Model(id="ACT-R Base")
    mod_graph = Graph(id="actr_base")
    mod.graphs.append(mod_graph)

    # Declarative memory
    dm_node = Node(id="declarative_memory")
    dm_node.parameters.append(Parameter(id="chunks", value=[]))
    dm_node.parameters.append(Parameter(id="chunk_types", value=[]))
    dm_ip = InputPort(id="dm_input")
    dm_node.input_ports.append(dm_ip)
    retrieval_f = Parameter(
        id="retrieve_chunk",
        function="retrieve_chunk",
        args={"pattern": dm_ip.id, "dm_chunks": "chunks", "types": "chunk_types"},
    )
    dm_node.parameters.append(retrieval_f)
    dm_op = OutputPort(id="dm_output", value=retrieval_f.id)
    dm_node.output_ports.append(dm_op)
    mod_graph.nodes.append(dm_node)

    # Retrieval buffer
    retrieval_node = Node(id="retrieval_buffer")
    retrieval_ip = InputPort(id="retrieval_input")
    retrieval_node.input_ports.append(retrieval_ip)
    retrieval_op = OutputPort(id="retrieval_output", value=retrieval_ip.id)
    retrieval_node.output_ports.append(retrieval_op)
    mod_graph.nodes.append(retrieval_node)

    # Goal buffer with state
    goal_node = Node(id="goal_buffer")
    goal_node.parameters.append(Parameter(id="first_goal", value={}))
    goal_ip = InputPort(id="goal_input")
    goal_node.input_ports.append(goal_ip)
    goal_f = Parameter(
        id="change_goal",
        function="change_goal",
        args={"pattern": goal_ip.id, "curr_goal": "goal_state"},
    )
    goal_node.parameters.append(goal_f)
    goal_state = Parameter(
        id="goal_state",
        default_initial_value="first_goal",
        value=f"first_goal if {goal_ip.id} == {{}} else {goal_f.id}",
    )
    goal_node.parameters.append(goal_state)
    goal_op = OutputPort(id="goal_output", value=goal_state.id)
    goal_node.output_ports.append(goal_op)
    mod_graph.nodes.append(goal_node)

    # Procedural memory
    pm_node = Node(id="procedural_memory")
    pm_node.parameters.append(Parameter(id="productions", value=[]))
    pm_op = OutputPort(id="pm_output", value="productions")
    pm_node.output_ports.append(pm_op)
    mod_graph.nodes.append(pm_node)

    # Pattern matching
    pattern_node = Node(id="pattern_matching")
    pattern_ip1 = InputPort(id="pattern_input_from_pm")
    pattern_ip2 = InputPort(id="pattern_input_from_goal")
    pattern_ip3 = InputPort(id="pattern_input_from_retrieval")
    pattern_node.input_ports.extend([pattern_ip1, pattern_ip2, pattern_ip3])
    pattern_f = Parameter(
        id="pattern_matching_function",
        function="pattern_matching_function",
        args={
            "productions": pattern_ip1.id,
            "goal": pattern_ip2.id,
            "retrieval": pattern_ip3.id,
        },
    )
    pattern_node.parameters.append(pattern_f)
    pattern_op = OutputPort(id="pattern_output", value=pattern_f.id)
    pattern_node.output_ports.append(pattern_op)
    mod_graph.nodes.append(pattern_node)

    # Conflict resolution
    conflict_node = Node(id="conflict_resolution")
    conflict_ip = InputPort(id="conflict_input")
    conflict_node.input_ports.append(conflict_ip)
    conflict_f = Parameter(
        id="conflict_resolution_function",
        function="conflict_resolution_function",
        args={"productions": conflict_ip.id},
    )
    conflict_node.parameters.append(conflict_f)
    conflict_op1 = OutputPort(id="conflict_output_to_fire_prod", value=conflict_f.id)
    conflict_op2 = OutputPort(id="conflict_output_to_check", value=conflict_f.id)
    conflict_node.output_ports.extend([conflict_op1, conflict_op2])
    mod_graph.nodes.append(conflict_node)

    # Node for firing productions
    fire_prod_node = Node(id="fire_production")
    fire_prod_ip = InputPort(id="fire_prod_input")
    fire_prod_node.input_ports.append(fire_prod_ip)
    fire_prod_f1 = Parameter(
        id="update_goal", function="update_goal", args={"production": fire_prod_ip.id}
    )
    fire_prod_f2 = Parameter(
        id="update_retrieval",
        function="update_retrieval",
        args={"production": fire_prod_ip.id},
    )
    fire_prod_node.parameters.extend([fire_prod_f1, fire_prod_f2])
    fire_prod_op1 = OutputPort(id="fire_prod_output_to_goal", value=fire_prod_f1.id)
    fire_prod_op2 = OutputPort(
        id="fire_prod_output_to_retrieval", value=fire_prod_f2.id
    )
    fire_prod_node.output_ports.extend([fire_prod_op1, fire_prod_op2])
    mod_graph.nodes.append(fire_prod_node)

    # Node to check termination
    check_node = Node(id="check_termination")
    check_ip = InputPort(id="check_input")
    check_node.input_ports.append(check_ip)
    check_f = Parameter(
        id="check_termination",
        function="check_termination",
        args={"production": check_ip.id},
    )
    check_node.parameters.append(check_f)
    check_op = OutputPort(id="check_output", value=check_f.id)
    check_node.output_ports.append(check_op)
    mod_graph.nodes.append(check_node)

    # Edges
    dm_to_retrieval = Edge(
        id="dm_to_pattern_edge",
        sender=dm_node.id,
        sender_port=dm_op.id,
        receiver=retrieval_node.id,
        receiver_port=retrieval_ip.id,
    )
    mod_graph.edges.append(dm_to_retrieval)

    retrieval_to_pattern = Edge(
        id="retrieval_to_pattern_edge",
        sender=retrieval_node.id,
        sender_port=retrieval_op.id,
        receiver=pattern_node.id,
        receiver_port=pattern_ip3.id,
    )
    mod_graph.edges.append(retrieval_to_pattern)

    goal_to_pattern = Edge(
        id="goal_to_pattern_edge",
        sender=goal_node.id,
        sender_port=goal_op.id,
        receiver=pattern_node.id,
        receiver_port=pattern_ip2.id,
    )
    mod_graph.edges.append(goal_to_pattern)

    pm_to_pattern = Edge(
        id="pm_to_pattern_edge",
        sender=pm_node.id,
        sender_port=pm_op.id,
        receiver=pattern_node.id,
        receiver_port=pattern_ip1.id,
    )
    mod_graph.edges.append(pm_to_pattern)

    pattern_to_conflict = Edge(
        id="pattern_to_conflict_edge",
        sender=pattern_node.id,
        sender_port=pattern_op.id,
        receiver=conflict_node.id,
        receiver_port=conflict_ip.id,
    )
    mod_graph.edges.append(pattern_to_conflict)

    conflict_to_fire_prod = Edge(
        id="conflict_to_fire_prod_edge",
        sender=conflict_node.id,
        sender_port=conflict_op1.id,
        receiver=fire_prod_node.id,
        receiver_port=fire_prod_ip.id,
    )
    mod_graph.edges.append(conflict_to_fire_prod)

    conflict_to_check = Edge(
        id="conflict_to_check_edge",
        sender=conflict_node.id,
        sender_port=conflict_op1.id,
        receiver=check_node.id,
        receiver_port=check_ip.id,
    )
    mod_graph.edges.append(conflict_to_check)

    # Conditions
    cond_dm = Condition(type="Always")
    cond_retrieval = Condition(type="JustRan", dependencies=dm_node.id)
    cond_goal = Condition(type="Always")
    cond_pm = Condition(type="Always")
    cond_pattern = Condition(
        type="And",
        dependencies=[
            Condition(type="EveryNCalls", dependencies=retrieval_node.id, n=1),
            Condition(type="EveryNCalls", dependencies=goal_node.id, n=1),
            Condition(type="EveryNCalls", dependencies=dm_node.id, n=1),
        ],
    )
    cond_conflict = Condition(type="JustRan", dependencies=pattern_node.id)
    cond_fire_prod = Condition(type="JustRan", dependencies=conflict_node.id)
    cond_check = Condition(type="JustRan", dependencies=conflict_node.id)
    cond_term = Condition(type="JustRan", dependencies=[check_node.id])
    mod_graph.conditions = ConditionSet(
        node_specific={
            dm_node.id: cond_dm,
            retrieval_node.id: cond_retrieval,
            goal_node.id: cond_goal,
            pm_node.id: cond_pm,
            pattern_node.id: cond_pattern,
            conflict_node.id: cond_conflict,
            fire_prod_node.id: cond_fire_prod,
            check_node.id: cond_check,
        },
        termination={"check_term_true": cond_term},
    )

    return mod


def actr_to_mdf(file_name: str):
    """Parses an ACT-R .lisp model file and outputs MDF .json and .yaml files.

    Args:
        file_name: The name of the ACT-R model file ending in .lisp.
    """
    with open(file_name) as actr_file:
        mod = build_model()
        add_dm = False
        add_lhs = False
        add_rhs = False
        chunk_types = {}
        dm = []
        pm = []
        goal = None
        curr_prod = None
        curr_pattern = None
        for l in actr_file:
            line = l.strip()
            # Add the chunk types
            if line.startswith("(chunk-type"):
                tokens = line.replace("(", "").replace(")", "").split(" ")
                chunk_types[tokens[1]] = tokens[2:]
            # Add the chunks to declarative memory
            elif line.startswith("(add-dm"):
                add_dm = True
            elif add_dm:
                if line == "":
                    add_dm = False
                else:
                    chunk = {}
                    tokens = line.replace("(", "").replace(")", "").split(" ")
                    chunk["name"] = tokens[0]
                    for i in range(1, len(tokens), 2):
                        chunk[tokens[i]] = tokens[i + 1]
                    dm.append(chunk)
            # Add the goal to the goal buffer
            elif line.startswith("(goal-focus"):
                goal_name = line[12:-1]
                for chunk in dm:
                    if chunk["name"] == goal_name:
                        goal = chunk
                        break
                for k in chunk_types[goal["ISA"]]:
                    if k not in goal.keys():
                        goal[k] = "nil"
            # Add the productions to procedural memory
            elif line.startswith(("(P", "(p")):
                curr_prod = {"name": line[3:], "lhs": [], "rhs": []}
                pm.append(curr_prod)
                add_lhs = True
            # Left hand side of production
            elif add_lhs:
                if line == "==>":
                    add_lhs = False
                    add_rhs = True
                elif line.endswith(">"):
                    curr_pattern = {"buffer": line[1:-1]}
                    curr_prod["lhs"].append(curr_pattern)
                else:
                    tokens = [t for t in line.split(" ") if t != ""]
                    if tokens[0] == "-":
                        curr_pattern[tokens[1]] = "-" + tokens[2]
                    else:
                        curr_pattern[tokens[0]] = tokens[1]
            # Right hand side of production
            elif add_rhs:
                if line == "" or line == ")" or "!output!" in line:
                    add_rhs = False
                elif line.endswith(">"):
                    curr_pattern = {"buffer": line[1:-1]}
                    curr_prod["rhs"].append(curr_pattern)
                else:
                    tokens = [t for t in line.split(" ") if t != ""]
                    if len(tokens) == 3:
                        curr_pattern[tokens[0] + tokens[1]] = tokens[2]
                    else:
                        curr_pattern[tokens[0]] = tokens[1]

        # Update production patterns
        for prod in pm:
            lhs = []
            for pattern in prod["lhs"]:
                isa = pattern["ISA"]
                final_pattern = {"buffer": pattern["buffer"], "ISA": isa}
                for k in chunk_types[isa]:
                    if k in pattern.keys():
                        final_pattern[k] = pattern[k]
                    else:
                        final_pattern[k] = "=" + k
                lhs.append(final_pattern)
            prod["lhs"] = lhs

        # Generate MDF files
        mod.id = (
            file_name[file_name.rindex("/") + 1 : -5]
            if "/" in file_name
            else file_name[:-5]
        )
        mod.graphs[0].id = mod.id + "_graph"
        mod.graphs[0].get_node("declarative_memory").get_parameter("chunks").value = dm
        mod.graphs[0].get_node("declarative_memory").get_parameter(
            "chunk_types"
        ).value = chunk_types
        mod.graphs[0].get_node("procedural_memory").get_parameter(
            "productions"
        ).value = pm
        mod.graphs[0].get_node("goal_buffer").get_parameter("first_goal").value = goal
        mod.to_json_file("%s.json" % file_name[:-5])
        mod.to_yaml_file("%s.yaml" % file_name[:-5])

        return mod



#!/usr/bin/env python

from setuptools import setup

# Most of this packages settings are defined in setup.cfg
# FIXME: Not sure of the best way to setup extras_require from setup.cfg
extras = {
    "test": [
        "pytest",
        "pytest-benchmark",
        "pytest-mock",
        "typing_extensions; python_version<'3.8'",
    ],
    "docs": [
        "Sphinx~=3.0",
        "recommonmark>=0.5.0",
        "nbsphinx",
        "sphinx_copybutton",
        "sphinx-rtd-theme",
        "myst_parser",
        "sphinx_markdown_tables",
        "sphinx-autoapi",
        "pytorch-sphinx-theme==0.0.19",
        "sphinxcontrib-versioning",
        "Jinja2<3.1",
    ],
    "dev": [],
}
extras["all"] = sum(extras.values(), [])
extras["dev"] += extras["test"]

setup(extras_require=extras)



"""
Implementation of core MDF function ontology.

This module implements and registers all builtin MDF functions.

"""
from typing import List, Dict

# Make sure we import math and numpy for Python expression strings. These imports
# are important, do not remove even though they appear unused.
import math
import numpy


"""
A dict that stores all registered MDF functions.
"""
mdf_functions = {}


def add_mdf_function(
    name: str = None,
    description: str = None,
    arguments: List[str] = None,
    expression_string: str = None,
):

    """Register a function with MDF function ontology.

    Adds a function to the registered list of available MDF functions.

    Args:
        name: name of the function e.g.'sin','cos','linear'
        description: Information about the function
        arguments: Inputs provided to obtain the result of function
        expression_string: Function expression in string format

    Returns:
        Updates mdf_functions

    """

    mdf_functions[name] = {}

    mdf_functions[name]["description"] = description
    mdf_functions[name]["arguments"] = arguments
    mdf_functions[name]["expression_string"] = expression_string
    try:
        mdf_functions[name]["function"] = create_python_function(
            name, expression_string, arguments
        )
    except SyntaxError:
        # invalid syntax in some onnx functions (e.g. onnx_ops.or)
        mdf_functions[name]["function"] = None


def create_python_expression(expression_string: str = None) -> str:
    """Converts the mathematical representation of function into function expression in python

    Args:
        expression_string: Mathematical expression of function in string format

    Returns:
        function expression in python
    """

    for func in ["exp", "sin", "cos", "tan", "sinh", "cosh", "tanh"]:
        if "numpy." + func not in expression_string:

            expression_string = expression_string.replace(
                "%s(" % func, "numpy.%s(" % func
            )
    for func in ["maximum"]:
        expression_string = expression_string.replace("%s(" % func, "numpy.%s(" % func)
    """for func in ["max"]:
        expression_string = expression_string.replace(
            "%s(" % func, "numpy.%simum(" % func
        )"""

    return expression_string


def substitute_args(expression_string: str = None, args: Dict[str, str] = None) -> str:
    """Substitute arg with the value in args dict

    Args:
        expression_string: function expression
        args: Dictionary of arguments
    Returns:
        modified expression string after substitution

    """
    # TODO, better checks for string replacement
    for arg in args:
        expression_string = expression_string.replace(arg, str(args[arg]))
    return expression_string


def create_python_function(
    name: str = None,
    expression_string: str = None,
    arguments: List[str] = None,
) -> "types.FunctionType":
    """Create a Python function e.g. linear, exponential, sin, cos, ReLu

    Args:
        name: name of the function e.g.'sin','cos','linear'
        expression_string: Function expression in string format
        arguments: list of inputs provided to obtain result from the function


    Returns:
        A function object

    """

    # assumes expression is one line
    name = name.replace(":", "_")
    expr = create_python_expression(expression_string)
    func_str = f"def {name}({','.join(arguments)}):\n\treturn {expr}"

    res = {}
    exec(func_str, globals(), res)
    return res[name]


# Populate the list of known functions

if len(mdf_functions) == 0:

    STANDARD_ARG_0 = "variable0"
    STANDARD_ARG_1 = "variable1"

    add_mdf_function(
        "linear",
        description="A linear function, calculated from a slope and an intercept",
        arguments=[STANDARD_ARG_0, "slope", "intercept"],
        expression_string="(%s * slope + intercept)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "logistic",
        description="Logistic function",
        arguments=[STANDARD_ARG_0, "gain", "bias", "offset"],
        expression_string="1/(1 + exp(-1*gain*(%s + bias) + offset))"
        % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "exponential",
        description="Exponential function",
        arguments=[STANDARD_ARG_0, "scale", "rate", "bias", "offset"],
        expression_string="scale * exp((rate * %s) + bias) + offset" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "sin",
        description="Sine function",
        arguments=[STANDARD_ARG_0, "scale"],
        expression_string="scale * sin(%s)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "cos",
        description="Cosine function",
        arguments=[STANDARD_ARG_0, "scale"],
        expression_string="scale * cos(%s)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "tan",
        description="Tangent function",
        arguments=[STANDARD_ARG_0, "scale"],
        expression_string="scale * tan(%s)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "sinh",
        description="Hyperbolic sine function",
        arguments=[STANDARD_ARG_0, "scale"],
        expression_string="scale * sinh(%s)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "cosh",
        description="Hyperbolic cosine function",
        arguments=[STANDARD_ARG_0, "scale"],
        expression_string="scale * cosh(%s)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "tanh",
        description="Hyperbolic tangent function",
        arguments=[STANDARD_ARG_0, "scale"],
        expression_string="scale * tanh(%s)" % (STANDARD_ARG_0),
    )

    add_mdf_function(
        "MatMul",
        description="Matrix multiplication (work in progress...)",
        arguments=["A", "B"],
        expression_string="A @ B",
    )

    add_mdf_function(
        "Relu",
        description="Rectified linear function (work in progress...)",
        arguments=["A"],
        expression_string="A * (A > 0)",
    )

    # Enumerate all available ONNX operators and add them as MDF functions.
    from modeci_mdf.functions.onnx import get_onnx_ops

    for mdf_func_spec in get_onnx_ops():
        add_mdf_function(**mdf_func_spec)

    # Add the ACT-R functions.
    from modeci_mdf.functions.actr import get_actr_functions

    for mdf_func_spec in get_actr_functions():
        add_mdf_function(**mdf_func_spec)


if __name__ == "__main__":

    import pprint

    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(mdf_functions)



from . import scheduler
from . import logger
import random
import inspect
import copy

# import ccm.config as config


class MethodWrapper:
    def __init__(self, obj, func, name):
        self.func = func
        self.obj = obj
        self.func_name = name
        self.begins = scheduler.Trigger(name + " begin")
        self.ends = scheduler.Trigger(name + " end")
        self.default_trigger = self.ends

    def __call__(self, *args, **keys):
        self.obj.sch.trigger(self.begins)
        val = self.func(self.obj, *args, **keys)
        self.obj.sch.trigger(self.ends)
        return val


class MethodGeneratorWrapper(MethodWrapper):
    def _generator(self, *args, **keys):
        self.obj.sch.trigger(self.begins)
        yield from self.func(self.obj, *args, **keys)
        self.obj.sch.trigger(self.ends)

    def __call__(self, *args, **keys):
        return self.obj.sch.add(self._generator, args=args, keys=keys)

    def __str__(self):
        return f"<MGW {self.obj} {self.func_name}>"


def log_everything(model, log=None):
    if log is None:
        log = logger.log_proxy
    if not hasattr(model, "log"):
        model.run(limit=0)
    model.log = log
    for k, v in model.__dict__.items():
        if k[0] != "_" and k != "parent":
            if isinstance(v, Model) and v.parent is model:
                log_everything(v, getattr(log, k))


class Model:
    __converted = False
    _convert_methods = True
    _auto_run_start = True
    name = "top"

    def __init__(self, log=None, **keys):
        self.__init_log = log
        for k, v in keys.items():
            setattr(self, k, v)

    def __convert(self, parent=None, name=None):
        # if self.__converted: return
        assert self.__converted == False
        self.__converted = True
        self.changes = scheduler.Trigger()

        if hasattr(self, "parent"):
            parent = self.parent

        methods = {}
        objects = {}
        for klass in inspect.getmro(self.__class__):
            if klass is not Model:
                for k, v in inspect.getmembers(klass):
                    if k[0] != "_":
                        if inspect.ismethod(v):
                            if (
                                k not in ["run", "now", "get_children"]
                                and k not in methods
                                and klass is not Model
                            ):
                                methods[k] = v
                        else:
                            if inspect.isclass(v) and Model in inspect.getmro(v):
                                v = v()
                            if k not in objects:
                                objects[k] = v
        objects = copy.deepcopy(objects)

        if parent:
            if not parent.__converted:
                parent.__convert()
            self.sch = parent.sch
            self.log = logger.dummy
            self.random = parent.random
            self.parent = parent
        else:
            self.sch = scheduler.Scheduler()
            if self.__init_log is True:
                self.log = logger.log()
            elif self.__init_log is None:
                self.log = logger.dummy
            else:
                self.log = self.__init_log
            self.random = random.Random()
            # seed=config.getOptions().random
            # if seed is not None: self.random.seed(seed)

            self.parent = None

        self._convert_info(objects, methods)

        for name, obj in objects.items():
            if isinstance(obj, Model):
                if not obj.__converted:
                    obj.__convert(self, name)
                else:
                    obj.name = name
                try:
                    self._children[name] = obj
                except AttributeError:
                    self._children = {name: obj}
            self.__dict__[name] = obj

        if self._convert_methods:
            for name, func in methods.items():
                if func.im_func.func_code.co_flags & 0x20 == 0x20:
                    w = MethodGeneratorWrapper(self, func, name)
                else:
                    w = MethodWrapper(self, func, name)
                self.__dict__[name] = w

        if self._auto_run_start:
            self.start()

        for k, v in self.__dict__.items():
            if k[0] != "_" and k != "parent" and isinstance(v, Model):
                if not v.__converted:
                    v.__convert(parent=self)

    def _convert_info(self, objects, methods):
        pass

    def __setattr__(self, key, value):
        if (
            key == "parent"
            and value is None
            and getattr(self, "parent", None) is not None
        ):
            del self.parent._children[self.name]
        self.__dict__[key] = value

        if isinstance(value, Model) and key[0] != "_" and key != "parent":
            self._ensure_converted()
            p = self
            ancestor = True
            while p is not None:
                if value is p:
                    break
                p = getattr(p, "parent", None)
            else:
                if getattr(value, "parent", None) is not None:
                    pass
                    # if value.name in value.parent._children:
                    #  del value.parent._children[value.name]
                else:
                    value.parent = self
                    value.name = key
                    try:
                        self._children[key] = value
                    except AttributeError:
                        self._children = {key: value}
                    if self.__converted and not value.__converted:
                        value.__convert(name=key, parent=self)

        if (
            self.__converted
            and key[0] != "_"
            and key not in ["parent", "sch", "changes", "log", "random", "name"]
        ):
            m = self
            done = []
            while m is not None:
                self.sch.trigger(m.changes, priority=-1)
                done.append(m)
                m = m.parent
                if m in done:
                    m = None
            if self.log:
                setattr(self.log, key, value)

        if key == "log" and value is not None:
            for k, v in self.__dict__.items():
                if k[0] != "_" and k not in [
                    "parent",
                    "sch",
                    "changes",
                    "log",
                    "random",
                    "name",
                ]:
                    if isinstance(v, (int, str, float, type(None))):
                        setattr(value, k, v)

    def start(self):
        pass

    def run(self, limit=None, func=None):
        if not self.__converted:
            self.__convert()
        # if config.getOptions().logall: log_everything(self)
        if limit is not None:
            self.sch.add(self.sch.stop, limit, priority=-9999999)
        if func is not None:
            self.sch.add(func)
        self.sch.run()

    def stop(self):
        if not self.__converted:
            self.__convert()
        self.sch.stop()

    def now(self):
        if not self.__converted:
            self.__convert()
        return self.sch.time

    def get_children(self):
        try:
            return self._children.values()
        except AttributeError:
            return []

    def _get_scheduler(self):
        self._ensure_converted()
        return self.sch

    def _ensure_converted(self):
        if not self.__converted:
            self.__convert()

    def _is_converted(self):
        return self.__converted



"""
Programmatically defines every ONNX operation as a python callable function. Executing ONNX graphs in this
way somewhat defeats the performance purposes of ONNX since the overhead for each operation will be high. However, this
allows us to test the MDF scheduler (which invokes Python functions) on any MDF model defined over ONNX operations. In
the future, the MDF should probably just compile to ONNX (or some other IR) for execution.
"""
import functools

import torch
import numpy as np
import onnxruntime as ort
import onnx.defs

# Currently using sklearn2onnx API to define ONNX operations. This dependency can probably be removed pretty easily.
# Do not remove this import even though it appears unused.
import skl2onnx.algebra.onnx_ops

from typing import Dict, Tuple, Any, List, Callable

OpSchema = onnx.defs.OpSchema
FormalParameterOption = OpSchema.FormalParameterOption

# Use the same ONNX opset version that torch is using for defaults now
# from torch.onnx.symbolic_helper import _default_onnx_opset_version as onnx_opset_version
onnx_opset_version = 13

__all__ = [
    "predict_with_onnxruntime",
    "run_onnx_op",
    "get_onnx_ops",
    "get_all_schemas_version",
]


def import_class(name: str) -> Any:
    """Import from a module specified by a string"""
    components = name.split(".")
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def predict_with_onnxruntime(model_def, *inputs) -> Dict[str, np.array]:
    """
    Simple helper to run an ONNX model with a set of inputs.

    Args:
        model_def: The ONNX model to run.
        *inputs: Input values to pass to the model.

    Returns:
        A dict of output values, keys are output names for the model. Values are
        the output values of the model.
    """
    sess = ort.InferenceSession(model_def.SerializeToString())
    names = [i.name for i in sess.get_inputs()]
    dinputs = {name: input for name, input in zip(names, inputs)}
    res = sess.run(None, dinputs)
    names = [o.name for o in sess.get_outputs()]
    return {name: output for name, output in zip(names, res)}


def convert_type(v):
    """Helper function to convert types to ONNX compatible types."""

    if type(v) == list:
        v = np.array(v)

    if type(v) == int or type(v) == float:
        v = np.atleast_1d(v)

    if hasattr(v, "dtype") and v.dtype == np.int32:
        v = v.astype(np.int64)

    if hasattr(v, "dtype") and v.dtype == np.float64:
        v = v.astype(np.float32)

    if isinstance(v, torch.Tensor):
        v = v.detach().cpu().numpy()

    return v


def run_onnx_op(
    op_name: str,
    inputs: Dict[str, np.array],
    output_names: List[str],
    opset_version: int = onnx_opset_version,
    **attributes,
):
    """
    Simple helper function that invokes a single ONNX operator with
    inputs and attibutes and returns the results. This isn't typically done
    in ONNX because graphs usually consist of more than one operation.
    This wrapper probably creates a significant amount of overhead for
    but if we want to execute and ONNX graph op by op it is the easiest
    thing to do.

    Args:
        op_name: The name of the operation to run, (Conv, Pad, etc.)
        inputs: A dict keyed by input name where the values are the input values to pass to the operation.
        output_names: The names to use for the output values.
        **attributes: Any additional attributes for the ONNX operation.

    Returns:
        A dict of output values, keys are output_names. Values are
        the output values of the operation.
    """

    # If the op name has the onnx namespace prefix, remove it.
    if "onnx::" in op_name:
        op_name = op_name.split("::")[-1]

    inputs = {k: convert_type(v) for k, v in inputs.items()}

    # In the case of the Pad operator, if constant_value argument is passed with different dtype
    # to the data, cast it.
    if op_name == "Pad":
        if "constant_value" in inputs:
            cval = inputs["constant_value"]
            data = list(inputs.values())[0]
            if cval.dtype != data.dtype:
                inputs["constant_value"] = cval.astype(data.dtype)

    # SkLearn ONNX doesn't seem to support ConcatFromSequence, see
    # https://github.com/onnx/sklearn-onnx/issues/710
    # Let us just use this implemetation I found in ONNX backend test.
    if op_name == "ConcatFromSequence":
        from onnx.backend.test.case.model.sequence import ConcatFromSequenceImpl

        return {
            "concat_result": ConcatFromSequenceImpl(
                inputs["input_sequence"], **attributes
            )
        }

    op_class = import_class(f"skl2onnx.algebra.onnx_ops.Onnx{op_name}")
    input_names = list(inputs.keys())
    input_vals = list(inputs.values())
    op = op_class(
        *input_names, output_names=output_names, op_version=opset_version, **attributes
    )

    model_def = op.to_onnx(inputs)
    return predict_with_onnxruntime(model_def, *input_vals)


def get_all_schemas_version(max_version: int) -> List[onnx.defs.OpSchema]:
    """
    Enumerate all the OpSchemas available from ONNX.

    Args:
        max_version: Only include up to max_version.

    Returns:
        A list of all OpSchemas
    """
    schemas = {}

    for schema in onnx.defs.get_all_schemas_with_history():
        # Only get ONNX ops available for this version
        if schema.domain == "" and schema.since_version <= max_version:
            if (
                schema.name in schemas
                and schemas[schema.name].since_version < schema.since_version
            ):
                schemas[schema.name] = schema
            elif schema.name not in schemas:
                schemas[schema.name] = schema

    return list(schemas.values())


@functools.lru_cache()
def get_onnx_schema(
    func_name: str, opset_version: int = onnx_opset_version
) -> onnx.defs.OpSchema:
    """Return the ONNX schema corresponding to a generated ONNX python
    function with name **func_name**

    Args:
        func_name (str): the name of the ONNX python function
        opset_version (int, optional): The opset version to use. Defaults to onnx_opset_version.

    Raises:
        ValueError: **func_name** does not correspond to a generated ONNX python function

    Returns:
        onnx.defs.OpSchema: The ONNX schema corresponding to function **func_name**
    """
    try:
        schema = list(
            filter(
                lambda f: f.name.lower() == func_name,
                get_all_schemas_version(onnx_opset_version),
            )
        )[0]
    except IndexError:
        raise ValueError("No corresponding onnx schema for %s" % func_name)
    else:
        return schema


def get_onnx_ops(opset_version: int = onnx_opset_version) -> List[Dict]:
    """
    Enumerate all available ONNX operations and generate MDF function specifications for each one.

    Args:
        opset_version: The opset version to use.

    Returns:
        A list of MDF function specifications. Each entry is a Dict that is feed directly to
            add_mdf_function.
    """

    mdf_funcspecs = []
    for schema in get_all_schemas_version(opset_version):
        args_list = [input.name for input in schema.inputs]
        params_list = [p for p in schema.attributes]
        args_params_str = ", ".join(args_list + params_list)

        mdf_funcspecs.append(
            dict(
                name=f"onnx::{schema.name}",
                description=schema.doc,
                arguments=args_list,
                expression_string=f"onnx_ops.{schema.name.lower()}({args_params_str})",
            )
        )

    return mdf_funcspecs


def _make_onnx_function(schema: onnx.defs.OpSchema) -> Callable:
    """
    Create and Python callable function from an ONNX OpSchema

    Args:
        schema: The ONNX schema to make a function from.

    Returns:

    """

    def onnx_wrapper(*args, **kwargs):

        # If we are dealing with cosntant, just return it.
        if schema.name == "Constant":
            value = args[0] if len(args) > 0 else list(kwargs.values())[0]
            return value

        # We have a workaround for ConstantOfShape, for some reason shape inference fails
        # if the optional value is passed as a numpy array. Convert to a TensorProto
        elif schema.name == "ConstantOfShape":
            if "value" in kwargs:
                kwargs["value"] = onnx.numpy_helper.from_array(
                    convert_type(kwargs["value"])
                )
            elif len(args) > 1:
                args[1] = onnx.numpy_helper.from_array(convert_type(kwargs["value"]))

        input_names = [inp.name for inp in schema.inputs]

        inputs_dict = {}

        # If the first argument is varidic, then any kwargs that are not in attributes
        # or are not passed as kwargs should passed into the input dict
        if (
            len(schema.inputs) > 0
            and schema.inputs[0].option == FormalParameterOption.Variadic
        ):

            for i, arg in enumerate(args):
                inputs_dict[f"input{i}"] = arg

            remove_list = []
            for kw, arg in kwargs.items():
                if kw not in schema.attributes:
                    inputs_dict[kw] = arg
                    remove_list.append(kw)

            # Remove any kwarg we have assigned to input, it is not an attribute
            for kw in remove_list:
                del kwargs[kw]

        else:
            # First, check if kwargs contains any inputs
            for kw in input_names:
                try:
                    inputs_dict[kw] = kwargs[kw]
                except KeyError:
                    pass

            # Assign any input names that have not yet been assigned by kwargs the remaning positional args
            arg_i = 0
            for inp_name in input_names:
                if inp_name not in inputs_dict:
                    if arg_i < len(args) and arg_i < len(input_names):
                        inputs_dict[inp_name] = args[arg_i]
                        arg_i = arg_i + 1

            # Remove any input argument specified in kwargs
            for input_arg in inputs_dict:
                if input_arg in kwargs:
                    del kwargs[input_arg]

            # If we have any remaining args. Assume they are attributes and assign them in order
            attribute_i = 0
            schema_attributes = list(schema.attributes)
            for arg in [arg for arg in args[arg_i:]]:
                while attribute_i < len(schema.attributes):
                    att_name = schema_attributes[attribute_i]

                    # If this attribute is already in kwargs, skip it
                    if att_name in kwargs:
                        attribute_i = attribute_i + 1
                    else:
                        kwargs[att_name] = arg
                        break

        # Check to make sure all the remaining kwargs are attributes
        for kw in kwargs:
            if kw not in schema.attributes:
                raise ValueError(
                    f"Passed unkown attribute ({kw}) to ONNX op {schema.name}, supported attributes: {list(schema.attributes)}"
                )

        output_names = [out.name for out in schema.outputs]

        out_dict = run_onnx_op(
            op_name=schema.name, inputs=inputs_dict, output_names=output_names, **kwargs
        )

        if len(out_dict) == 1:
            return tuple(out_dict.values())[0]
        else:
            return tuple(out_dict.values())

    return onnx_wrapper


def _define_onnx_functions(opset_version):
    """
    Enumerate all ONNX operators and define Python Callable functions for each one. This kind of defeats the purpose of
    ONNX since we are paying the overhead for invoking each of these functions separately from Python. However, for now,
    this hack will allow us to test any MDF model that is composed of ONNX functions through the scheduler.
    """
    import sys

    current_module = sys.modules[__name__]

    for schema in get_all_schemas_version(opset_version):

        onnx_wrapper = _make_onnx_function(schema)

        # Lets call this function a lowercase version of the opname, follow PEP 8
        func_name = schema.name.lower()

        # Lets add some documentation.
        onnx_wrapper.__doc__ = schema.doc

        setattr(current_module, func_name, onnx_wrapper)


# Define all ONNX operators as functions on this module.
_define_onnx_functions(onnx_opset_version)



"""Contains implementations of ACT-R functions using the ccm library."""

import random
from .ccm.pattern import Pattern
from .ccm.scheduler import Scheduler
from .ccm.dm import Memory
from .ccm.buffer import Chunk, Buffer
from typing import Union, Dict, Any, Tuple, List, Callable


def chunk_to_string(chunk: Dict[str, str]) -> str:
    """Converts a chunk dictionary to a string format.

    Args:
        chunk: A dict representing a chunk.

    Returns:
        A string representation of the chunk.
    """
    chunk_copy = chunk.copy()
    if "name" in chunk_copy.keys():
        del chunk_copy["name"]
    return " ".join(list(chunk_copy.values()))


def pattern_to_string(pattern: Dict[str, str]) -> str:
    """Converts a pattern dictionary to a string format.

    Args:
        chunk: A dict representing a pattern.

    Returns:
        A string representation of the pattern.
    """
    return " ".join(list(pattern.values())[1:]).replace("-=", "!?").replace("=", "?")


def change_goal(pattern: Dict[str, str], curr_goal: Dict[str, str]):
    """Modifies the current goal buffer using the given pattern.

    Args:
        pattern: A dict representing a pattern.
        curr_goal: A dict representing the current goal pattern.

    Returns:
        The current goal updated with the data in pattern.
    """
    if curr_goal == 0:
        return {}
    curr_goal.update(pattern)
    if "buffer" in curr_goal.keys():
        del curr_goal["buffer"]
    return curr_goal


def retrieve_chunk(
    pattern: Dict[str, str],
    dm_chunks: List[Dict[str, str]],
    types: Dict[str, List[str]],
) -> Dict[str, str]:
    """Retrieve a chunk from declarative memory given a pattern.

    Args:
        pattern: A dict representing the pattern to match.
        dm_chunks: A list of dicts, each representing a chunk in declarative memory.
        types: A dict containing each possible chunk type.

    Returns:
        The chunk in declarative memory that matches the pattern.
    """
    if pattern == {}:
        return {}
    retrieve = Buffer()
    memory = Memory(retrieve)
    for chunk in dm_chunks:
        memory.add(chunk_to_string(chunk))
    memory.sch = Scheduler()
    matches = memory.find_matching_chunks(pattern_to_string(pattern))
    if matches == []:
        return {}
    match = matches[0]
    isa = match[0]
    retrieved = {"ISA": isa}
    for i in range(1, len(match.values())):
        retrieved[types[isa][i - 1]] = match[i]
    return retrieved


def match_production(
    production: Dict[str, Any], context: Dict[str, Dict[str, str]]
) -> bool:
    """Returns True if the production's left hand side matches the given context
    and adds the matching bindings to the production.

    Args:
        production: A dict representing a production.
        context: A dict with the contents of the goal and retrieval buffers.

    Returns:
        True if the production's left hand side matches the context.
    """
    patterns = {}
    for p in production["lhs"]:
        patterns[p["buffer"]] = pattern_to_string(p)
    patt = Pattern(patterns)
    match_bindings = patt.match(context)
    if match_bindings is None:
        return False
    production["bindings"] = match_bindings
    return True


def pattern_matching_function(
    productions: List[Dict[str, Any]], goal: Dict[str, str], retrieval: Dict[str, str]
) -> List[Dict[str, Any]]:
    """Returns the productions that match the given goal and retrieval buffers.

    Args:
        productions: A list of all productions as dicts.
        goal: The current value of the goal buffer as a dict.
        retrieval: The chunk dict retrieved from declarative memory.

    Returns:
        A list of productions that match the buffers.
    """
    context = {
        "goal": Chunk(chunk_to_string(goal)),
        "retrieval": Chunk(chunk_to_string(retrieval)) if retrieval != {} else None,
    }
    return [p for p in productions if match_production(p, context)]


def conflict_resolution_function(productions: List[Dict[str, Any]]) -> Dict[str, Any]:
    """ACT-R conflict resolution function. Currently selects a production at
    random from the already matched productions, since utility values and learning
    are not implemented yet.

    Args:
        productions: A list of productions as dicts.

    Returns:
        The selected production from the list.
    """
    if len(productions) == 0:
        return {}
    else:
        return random.choice(productions)


def update_buffer(production: Dict[str, Any], buffer: str) -> Dict[str, str]:
    """Returns a pattern to update the given buffer with.

    Args:
        production: The production dict that specifies the buffer update.
        buffer: The name of the buffer to update.

    Returns:
        A pattern that the buffer will be updated with.
    """
    if len(production) == 0:
        return {}
    pattern = {}
    for p in production["rhs"]:
        if p["buffer"] == buffer:
            pattern = p.copy()
    for k, v in pattern.items():
        v_name = v.replace("=", "")
        if v_name in production["bindings"].keys():
            pattern[k] = production["bindings"][v_name]
    return pattern


def update_goal(production: Dict[str, Any]) -> Dict[str, str]:
    """Returns a pattern to update the goal buffer with.

    Args:
        production: The production dict that specifies the goal buffer update.

    Returns:
        A pattern that the goal buffer will be updated with.
    """
    return update_buffer(production, "goal")


def update_retrieval(production: Dict[str, Any]) -> Dict[str, str]:
    """Returns a pattern to update the retrieval buffer with.

    Args:
        production: The production dict that specifies the retrieval buffer update.

    Returns:
        A pattern that the retrieval buffer will be updated with.
    """
    return update_buffer(production, "retrieval")


def check_termination(production: Dict[str, Any]):
    """Function used to check if no production was selected.

    Args:
        production: A production dict that was selected.

    Returns:
        True if the production is empty.
    """
    return production == {}


def get_actr_functions() -> List[Dict[str, Any]]:
    """Creates a list of all the ACT-R functions as MDF specifications.

    Returns:
        A list of MDF function specifications.
    """
    actr_funcs = []
    actr_funcs.append(
        dict(
            name="change_goal",
            description="ACT-R change goal buffer function",
            arguments=["pattern", "curr_goal"],
            expression_string="actr.change_goal(pattern, curr_goal)",
        )
    )
    actr_funcs.append(
        dict(
            name="retrieve_chunk",
            description="ACT-R retrieve chunk function",
            arguments=["pattern", "dm_chunks", "types"],
            expression_string="actr.retrieve_chunk(pattern, dm_chunks, types)",
        )
    )
    actr_funcs.append(
        dict(
            name="pattern_matching_function",
            description="ACT-R pattern matching function",
            arguments=["productions", "goal", "retrieval"],
            expression_string="actr.pattern_matching_function(productions, goal, retrieval)",
        )
    )
    actr_funcs.append(
        dict(
            name="conflict_resolution_function",
            description="ACT-R conflict resolution function",
            arguments=["productions"],
            expression_string="actr.conflict_resolution_function(productions)",
        )
    )
    actr_funcs.append(
        dict(
            name="update_goal",
            description="ACT-R update goal buffer function",
            arguments=["production"],
            expression_string="actr.update_goal(production)",
        )
    )
    actr_funcs.append(
        dict(
            name="update_retrieval",
            description="ACT-R update retrieval buffer function",
            arguments=["production"],
            expression_string="actr.update_retrieval(production)",
        )
    )
    actr_funcs.append(
        dict(
            name="check_termination",
            description="check_termination",
            arguments=["production"],
            expression_string="actr.check_termination(production)",
        )
    )
    return actr_funcs
